/*
 *  liboms-pcix.c --- implementation of the OMS motion control library for
 *     Linux and the PCIx board.
 *
 *     Note that the library and device driver were developed in
 *     tandem --- the distribution of functionality between kernel-space
 *     and user-space may not correspond with that of Windows systems.
 *
 *     This file, and perhaps others in this directory (see Makefile)
 *     will be compiled and built into a shared library (".so").
 *
 *  Copyright (c) 1997 - 2003  Oregon Micro Systems.  All Rights Reserved.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <asm/types.h>   // need cpu-specific u32
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <sys/file.h>
#include <time.h>
#include <unistd.h>
#include <malloc.h>

#include "liboms-pcix.h"     // defines API of this library
#include "oms-pcix.h"        // defines driver's ioctl API

static int debuglevel;       // used for debugging

#define MAX_CARDS  16

// HANDLE -- rather than have OMS_GetHandle() return a raw file descriptor,
//   we return an opaque pointer to a structure containing the file
//   descriptor, plus anything else we need to manage a handle.
struct namemap {
        char *omsname;
        char *linuxname;
};

struct handle {
        int             fd;        // file descriptor
        struct namemap *namemap;   // names
};

static struct namemap namemap[] = {
        {"OmsPCIx1",  "/dev/oms-pcix0"}, {"OmsPCIx2",  "/dev/oms-pcix1"},
        {"OmsPCIx3",  "/dev/oms-pcix2"}, {"OmsPCIx4",  "/dev/oms-pcix3"},
        {"OmsPCIx5",  "/dev/oms-pcix4"}, {"OmsPCIx6",  "/dev/oms-pcix5"},
        {"OmsPCIx7",  "/dev/oms-pcix6"}, {"OmsPCIx8",  "/dev/oms-pcix7"},
        {"OmsPCIx9",  "/dev/oms-pcix8"}, {"OmsPCIx10", "/dev/oms-pcix9"},
        {"OmsPCIx11", "/dev/oms-pcix10"},{"OmsPCIx12", "/dev/oms-pcix11"},
        {"OmsPCIx13", "/dev/oms-pcix12"},{"OmsPCIx14", "/dev/oms-pcix13"},
        {"OmsPCIx15", "/dev/oms-pcix14"},{"OmsPCIx16", "/dev/oms-pcix15"},
        { NULL, NULL}
};

/*************************************************************************
 *                  Private (static) functions                           *
 *************************************************************************/

// validate_handle() --- perform rough validation that a HANDLE passed from
//   a user program is really a handle.  Returns TRUE if valid, FALSE if not.
static int
validate_handle(struct handle *h)
{
        int namemap_offset;

        if (h == NULL) {
                if (debuglevel)
                        fprintf(stderr, "validate_handle:  NULL handle\n");
                return 0;
        }
        if (h->fd < 0) {
                if (debuglevel)
                        fprintf(stderr, "validate_handle:  negative fd\n");
                return 0;
        }
        if (h->namemap == NULL) {
                if (debuglevel)
                        fprintf(stderr, "validate_handle:  NULL namemap\n");
                return 0;
        }
        // verify that the namemap field is in range of the namemap array
        namemap_offset = (unsigned long)(h->namemap) - (unsigned long)namemap;
        if (namemap_offset < 0 || namemap_offset >= sizeof(namemap)) {
                if (debuglevel)
                        fprintf(stderr, "validate_handle:  namemap offset "
                                "out of range\n");
                return 0;
        }
        // verify that the namemap field points to the beginning of one
        //  of the elements of the namemap array.
        if (namemap_offset % sizeof(struct namemap)) {
                if (debuglevel)
                        fprintf(stderr, "validate_handle:  namemap index "
                                "non-integer\n");
                return 0;
        }
        return 1;
}

// msec_delay() --- delay a specified number of millesecs.  Resumes
//   after handling any signals.
static void
msec_delay(int msecs)
{
        struct timespec req, rem;
        
        memset(&req, 0, sizeof(req));
        while (msecs >= 1000) {
                ++(req.tv_sec);
                msecs -= 1000;
        }
        req.tv_nsec = msecs * 1000000;
        while (nanosleep(&req, &rem)) {
                req.tv_sec = rem.tv_sec;
                req.tv_nsec = rem.tv_nsec;
        }
}

// get_single_axis_char() --- map bit (OMS_X_AXIS, etc.) to char ('X', etc.).
static char
get_single_axis_char(long axis_selection)
{
	switch (axis_selection) {
	case OMS_X_AXIS:  return 'X';
	case OMS_Y_AXIS:  return 'Y';
	case OMS_Z_AXIS:  return 'Z';
	case OMS_T_AXIS:  return 'T';
	case OMS_U_AXIS:  return 'U';
	case OMS_V_AXIS:  return 'V';
	case OMS_R_AXIS:  return 'R';
	case OMS_S_AXIS:  return 'S';
	default:          return 0;
	}
}

		
// send_query_command() --- issue the commonly used IOCTL_OMS_QUERY
//    Unfortunately, the only return values allowed are SUCCESS and
//    RESPONSE_TIME_OUT.  (Use 'debuglevel' for more info).
static long
send_query_command(OMSHANDLE handle, const char *cmd, char *response)
{
        struct oms_ioctl  ioc;
        struct handle     *h = handle;  // should be already validated

        ioc.command = (char *) cmd;
        ioc.command_len = strlen(cmd) + 1;
        ioc.response = response;
        ioc.response_len = 128;    // OMS docs say to allow this much space
        
        if (ioctl(h->fd, IOCTL_OMS_QUERY, (unsigned long) &ioc)) {
                if (debuglevel)
                        fprintf(stderr, "%s:  ioctl %s\n",
                                __FUNCTION__, strerror(errno));
                *response = 0;  // delimit user's string to empty
                return RESPONSE_TIME_OUT;
        }
        return SUCCESS;
}

/*  macro for consolidating common code validating handle */
#define V_HANDLE(h, returncode)                                             \
   ({                                                                       \
        if (!validate_handle(h)) {                                          \
                if (debuglevel)                                             \
                        fprintf(stderr, "%s:  bad handle\n", __FUNCTION__); \
		return returncode;                                          \
        }                                                                   \
    })

long
getflags(OMSHANDLE handle, long  *pflags, long ioctlcmd)
{
        struct handle     *h = handle;
        __u32              response;
        struct oms_ioctl  ioc;

	V_HANDLE(h, COMMAND_TIME_OUT);

        memset(&ioc, 0, sizeof(ioc));
        ioc.response = (char *) &response;
        ioc.response_len = sizeof(__u32);
        if (ioctl(h->fd, ioctlcmd, &ioc)) {
                if (debuglevel)
                        fprintf(stderr, "%s:  ioctl"
                                ":  %s\n", __FUNCTION__,
                                strerror(errno));
                        return INVALID_AXIS_SELECTION;  // only choice
                }
        *pflags = (response & 0xff);
        return SUCCESS;
        return INVALID_AXIS_SELECTION;
}

long
clrflags(OMSHANDLE handle, long flags, long ioctlcmd)
{
	struct handle     *h = handle;
	struct oms_ioctl  ioc;
	__u32              mask = flags;

	V_HANDLE(h, INVALID_AXIS_SELECTION);

	if (mask == 0 || (mask & 0xffffff00) != 0)
		return INVALID_AXIS_SELECTION;
	
	memset(&ioc, 0, sizeof(ioc));
	ioc.command = (unsigned char *) &mask;
	ioc.command_len = sizeof(__u32);
	if (ioctl(h->fd, ioctlcmd, &ioc)) {
		if (debuglevel)
			fprintf(stderr, "%s:  ioctl:  %s\n",
				__FUNCTION__, strerror(errno));
		return INVALID_AXIS_SELECTION;
	}
	return SUCCESS;
}

/*************************************************************************
 *                       Exported functions                              *
 *************************************************************************/

OMSHANDLE
OMS_GetHandle(const char *devname)
{
        struct namemap *nm;
        struct handle  *h;
        int             fd;
        
        for (nm = namemap; nm->omsname; nm++) {
                if (strcmp(nm->omsname, devname) == 0)
                        break;
        }
        if (nm->omsname == NULL) {
                if (debuglevel)
                        fprintf(stderr, "GetOmsHandle:  name \"%s\" invalid\n",
                                devname);
                return NULL;
        }
        if ((fd = open(nm->linuxname, O_RDWR)) < 0) {
                if (debuglevel)
                        fprintf(stderr, "%s:  open(2) %s:  %s\n",
				__FUNCTION__,
                                nm->linuxname, strerror(errno));
                return NULL;
        }
        if ((h = malloc(sizeof(struct handle))) == NULL) {
                close(fd);
                if (debuglevel)
                        fprintf(stderr, "GetOmsHandle:  malloc failed\n");
                return NULL;
        }
        h->fd = fd;
        h->namemap = nm;
        return h;
}

void
OMS_CloseHandle(OMSHANDLE handle)
{
        struct handle *h = handle;

	V_HANDLE(h, /*void*/);

        if (close(h->fd)) {
                if (debuglevel)
                        fprintf(stderr, "%s:  close(2) fd %d: %s\n",
				__FUNCTION__, h->fd, strerror(errno));
                return;
        }
        // make data in handle invalid before freeing
        h->fd = -1;
        h->namemap = NULL;
        free(h);
}

long
OMS_ResetController(OMSHANDLE handle)
{
        struct handle     *h = handle;
        struct oms_ioctl  ioc;
        char              tmp[128];

	V_HANDLE(h, COMMAND_TIME_OUT);

        memset(&ioc, 0, sizeof(ioc));
        ioc.command = "RS";
        ioc.command_len = 2;
        if (ioctl(h->fd, IOCTL_OMS_COMMAND, (unsigned long) &ioc)) {
                if (debuglevel)
                        fprintf(stderr, "%s:  OMS_COMMAND ioctl:  %s\n",
                                __FUNCTION__, strerror(errno));
                return COMMAND_TIME_OUT;
        }
	msec_delay(5000);  // allow the controller to reset
	return send_query_command(h, "WY", tmp);
}

long
OMS_GetControllerDescription(OMSHANDLE handle, char *description)
{
	struct handle  *h = handle;

	V_HANDLE(h, RESPONSE_TIME_OUT);
	return send_query_command(h, "WY", description);
}

long
OMS_SendAndGetString(OMSHANDLE handle, const char *cmd, char *response)
{
        msec_delay(5);  // enforce 5 millesec delay between queries, to
                        //  avoid overloading driver
	V_HANDLE(handle, RESPONSE_TIME_OUT);

        return send_query_command(handle, cmd, response);
}

long
OMS_SendString(OMSHANDLE handle, const char *cmd)
{
        struct handle     *h = handle;
        struct oms_ioctl  ioc;

	V_HANDLE(h, COMMAND_TIME_OUT);

        memset(&ioc, 0, sizeof(ioc));
        ioc.command = (char *) cmd;
        ioc.command_len = strlen(cmd) + 1;
        if (ioctl(h->fd, IOCTL_OMS_COMMAND, (unsigned long) &ioc)) {
                if (debuglevel)
                        fprintf(stderr, "%s:  OMS_COMMAND ioctl:  %s\n",
                                __FUNCTION__, strerror(errno));
                return COMMAND_TIME_OUT;
        }
        return SUCCESS;
}

long
OMS_GetCmdBufferFree(OMSHANDLE handle)
{
        struct handle     *h = handle;
        struct oms_ioctl  ioc;
        __u32              result = 0;

	V_HANDLE(h, 0);

        memset(&ioc, 0, sizeof(ioc));
        ioc.response = (char *) &result;
        ioc.response_len = sizeof(__u32);
        if (ioctl(h->fd, IOCTL_OMS_CMD_BUFF_FREE, (unsigned long) &ioc)) {
                if (debuglevel)
                        fprintf(stderr, "%s:  CMD_BUFF_FREE ioctl:"
                                "  %s\n", __FUNCTION__, strerror(errno));
                return 0;
        }
        return result;
}

long
OMS_Wait(long waittime)
{
	if ((waittime >= 0) && (waittime <= 3600000)) {
		msec_delay(waittime);
		return SUCCESS;
	}
	return INVALID_PARAMETER;
}

long
OMS_StopMotion(OMSHANDLE handle, long axis_selection)
{
	struct handle *h = handle;
	char           tmp[32], command[128];
	long           mask = OMS_X_AXIS;
	int            i;

	V_HANDLE(h, COMMAND_TIME_OUT);  // should be INVALID_PARAMETER
	command[0] = 0;
	for ( i = 0; i < 8; i++ )
	{
		sprintf(tmp, "AXST");
		mask = mask << i;
		if ( (mask & axis_selection) > 0 )
		{
			if ((tmp[1] = get_single_axis_char(mask)) == 0)
				return INVALID_AXIS_SELECTION;
			strcat(command, tmp);
		}
	}
	return OMS_SendString(h, command);
}

long
OMS_KillMotion(OMSHANDLE handle, long axis_selection)	
{
	struct handle *h = handle;
	char           tmp[32], command[128];
	long           mask = OMS_X_AXIS;
	int            i;

	V_HANDLE(h, COMMAND_TIME_OUT);  // should be INVALID_PARAMETER
	command[0] = 0;
	for ( i = 0; i < 8; i++ )
	{
		sprintf(tmp, "AXKL");
		mask = mask << i;
		if ( (mask & axis_selection) > 0 )
		{
			if ((tmp[1] = get_single_axis_char(mask)) == 0)
				return INVALID_AXIS_SELECTION;
			strcat(command, tmp);
		}
	}
	return OMS_SendString(h, command);
}

long
OMS_GetMotorPositions(OMSHANDLE handle,
                      long axis_selection,
                      PAXES_DATA ppos)
{
	struct handle     *h = handle;
	__u32              positions[16];
	struct oms_ioctl  ioc;
	__u32              as = axis_selection;

	if (as == 0 || as > OMS_ALL_AXES)
		return INVALID_AXIS_SELECTION;
	V_HANDLE(h, RESPONSE_TIME_OUT);
	memset(&ioc, 0, sizeof(ioc));
	ioc.response = (char *) positions;
	ioc.response_len = sizeof(positions);
	if (ioctl(h->fd, IOCTL_OMS_READ_AXIS_POSITIONS, &ioc)) {
		if (debuglevel)
			fprintf(stderr, "%s:  ioctl:  %s\n",
				__FUNCTION__, strerror(errno));
		return RESPONSE_TIME_OUT;  // only choice in spec
	}
	if (OMS_X_AXIS & as)
		ppos->X = positions[0];
	if (OMS_Y_AXIS & as)
		ppos->X = positions[1];
	if (OMS_Z_AXIS & as)
		ppos->X = positions[2];
	if (OMS_T_AXIS & as)
		ppos->X = positions[3];
	if (OMS_U_AXIS & as)
		ppos->X = positions[4];
	if (OMS_V_AXIS & as)
		ppos->X = positions[5];
	if (OMS_R_AXIS & as)
		ppos->X = positions[6];
	if (OMS_S_AXIS & as)
		ppos->X = positions[7];
	return SUCCESS;
}

long
OMS_GetEncoderPositions(OMSHANDLE handle,
                        long axis_selection,
                        PAXES_DATA ppos)
{
	struct handle     *h = handle;
	__u32              positions[16];
	struct oms_ioctl  ioc;
	__u32              as = axis_selection;

	if (as == 0 || as > OMS_ALL_AXES)
		return INVALID_AXIS_SELECTION;
	V_HANDLE(h, RESPONSE_TIME_OUT);
	memset(&ioc, 0, sizeof(ioc));
	ioc.response = (char *) positions;
	ioc.response_len = sizeof(positions);
	if (ioctl(h->fd, IOCTL_OMS_READ_AXIS_POSITIONS, &ioc)) {
		if (debuglevel)
			fprintf(stderr, "%s:  ioctl:  %s\n",
				__FUNCTION__, strerror(errno));
		return RESPONSE_TIME_OUT;  // only choice in spec
	}
	if (OMS_X_AXIS & as)
		ppos->X = positions[8];
	if (OMS_Y_AXIS & as)
		ppos->X = positions[9];
	if (OMS_Z_AXIS & as)
		ppos->X = positions[10];
	if (OMS_T_AXIS & as)
		ppos->X = positions[11];
	if (OMS_U_AXIS & as)
		ppos->X = positions[12];
	if (OMS_V_AXIS & as)
		ppos->X = positions[13];
	if (OMS_R_AXIS & as)
		ppos->X = positions[14];
	if (OMS_S_AXIS & as)
		ppos->X = positions[15];
	return SUCCESS;
}

long
OMS_GetPositionData(OMSHANDLE handle, PPOSITION_DATA pdat)
{
	struct oms_ioctl  ioc;
	struct handle     *h = handle;
	__u32              positions[16];

	V_HANDLE(h, COMMAND_TIME_OUT);  // should be INVALID_PARAMETER
	memset(&ioc, 0, sizeof(ioc));
	ioc.response = (char *) positions;
	ioc.response_len = sizeof(positions);
	if (ioctl(h->fd, IOCTL_OMS_READ_AXIS_POSITIONS, &ioc)) {
		if (debuglevel)
			fprintf(stderr, "%s:  ioctl:  %s\n",
				__FUNCTION__, strerror(errno));
		return COMMAND_TIME_OUT;  // only choice in spec
	}
	pdat->X.Motor   = positions[ 0];
	pdat->Y.Motor   = positions[ 1];
	pdat->Z.Motor   = positions[ 2];
	pdat->T.Motor   = positions[ 3];
	pdat->U.Motor   = positions[ 4];
	pdat->V.Motor   = positions[ 5];
	pdat->R.Motor   = positions[ 6];
	pdat->S.Motor   = positions[ 7];
	pdat->X.Encoder = positions[ 8];
	pdat->Y.Encoder = positions[ 9];
	pdat->Z.Encoder = positions[10];
	pdat->T.Encoder = positions[11];
	pdat->U.Encoder = positions[12];
	pdat->V.Encoder = positions[13];
	pdat->R.Encoder = positions[14];
	pdat->S.Encoder = positions[15];
	return SUCCESS;
}

long
OMS_GetControllerFlags(OMSHANDLE handle, long *pflags)
{
        return getflags(handle, pflags, IOCTL_OMS_READ_STATUS_FLAGS);
}

long
OMS_ClrControllerFlags(OMSHANDLE handle, long flags)
{
        return clrflags(handle, flags, IOCTL_OMS_CLEAR_STATUS_FLAGS);
}

long
OMS_GetDoneFlags(OMSHANDLE handle, long  *pflags)
{
        return getflags(handle, pflags, IOCTL_OMS_READ_DONE_FLAGS);
}

long
OMS_ClrDoneFlags(OMSHANDLE handle, long flags)
{
        return clrflags(handle, flags, IOCTL_OMS_CLEAR_DONE_FLAGS);
}

long
OMS_GetPosLimitFlags(OMSHANDLE handle, long  *pflags)
{
        return getflags(handle, pflags, IOCTL_OMS_READ_PLIMIT_FLAGS);
}

long
OMS_ClrPosLimitFlags(OMSHANDLE handle, long flags)
{
        return clrflags(handle, flags, IOCTL_OMS_CLEAR_PLIMIT_FLAGS);
}

long
OMS_GetNegLimitFlags(OMSHANDLE handle, long  *pflags)
{
        return getflags(handle, pflags, IOCTL_OMS_READ_NLIMIT_FLAGS);
}

long
OMS_ClrNegLimitFlags(OMSHANDLE handle, long flags)
{
        return clrflags(handle, flags, IOCTL_OMS_CLEAR_NLIMIT_FLAGS);
}

long
OMS_GetSlipFlags(OMSHANDLE handle, long  *pflags)
{
        return getflags(handle, pflags, IOCTL_OMS_READ_SLIP_FLAGS);
}

long
OMS_ClrSlipFlags(OMSHANDLE handle, long flags)
{
        return clrflags(handle, flags, IOCTL_OMS_CLEAR_SLIP_FLAGS);
}

long
GetOmsAxisFlags(OMSHANDLE handle, long axis_selection, long *pflags) 
{
	struct handle     *h = handle;
	char               command[5];
	char               response[128];
	long               tmpflags;
	long               flags = 0;
	long               status;

	V_HANDLE(h, INVALID_AXIS_SELECTION);

	strcpy(command, "AXQA");
	if ((command[1] = get_single_axis_char(axis_selection)) == 0)
		return INVALID_AXIS_SELECTION;
	if (OMS_GetSlipFlags(h, &tmpflags))
		return INVALID_AXIS_SELECTION;
	if (tmpflags & axis_selection)
		flags |= (1 << 1);   // bit 1
	OMS_GetPosLimitFlags(h, &tmpflags);
	if (tmpflags & axis_selection)
		flags |= (1 << 2);   // bit 2
        else {
		OMS_GetNegLimitFlags(h, &tmpflags);
		if (tmpflags & axis_selection)
			flags |= (1 << 2);   // bit 2
        }
	// send QA query to get the other axis flags
	if ((status = send_query_command(h, command, response)) != SUCCESS)
		return status;
	if (response[0] == 'M')
		flags |= (1 << 0);   // bit 0
	if (response[1] == 'H')
		flags |= (1 << 3);   // bit 4
	*pflags = flags;
	return SUCCESS;
}
	
long
GetOmsEncoderFlags(OMSHANDLE handle, long axis_selection, long *pflags)
{
	struct handle     *h = handle;
	char               command[5];
	char               response[128];
	long               tmpflags;
	long               flags = 0;
	long               status;

	V_HANDLE(h, INVALID_AXIS_SELECTION);

	strcpy(command, "AXEA");
	if ((command[1] = get_single_axis_char(axis_selection)) == 0)
		return INVALID_AXIS_SELECTION;
	status = OMS_GetSlipFlags(h, &tmpflags);
	if (status != SUCCESS)
		return status;
	if (tmpflags & axis_selection)
		flags |= (1 << 2);   // bit 2 SLIP OR STALL DETECTED
	status = send_query_command(h, command, response);
	if (status != SUCCESS)
		return status;
	// slip detect enable
	if (response[0] == 'E')
		flags |= (1 << 0);   // bit 0 SLIP DETECT ENABLE
	if (response[1] == 'E')
		flags |= (1 << 1);   // bit 1 POSITION MAINTENANCE ENABLE
	if (response[3] == 'P')
		flags |= (1 << 3);   // bit 3 WITHIN DEADBAND
	if (response[4] == 'H')
		flags |= (1 << 4);   // bit 4 AXIS HOME
	*pflags = flags;
	return SUCCESS;
}

/***************************************************************************
 *          UNDOCUMENTED exported functions (used for debugging)           *
 ***************************************************************************/

void
_omsdebug_set_debug_level(int lev)
{
        debuglevel = lev;
}

void
_omsdebug_test_handle_validation(void)
{
        struct handle h;
        int    i;
        unsigned long fudge;

        for (i = 0; i < MAX_CARDS; i++) {
                h.fd = i;
                h.namemap = namemap + i;
                if (!validate_handle(&h)) {
                        fprintf(stderr, "OOPS:  handle %d not valid\n", i);
                        return;
                } else {
                        fprintf(stderr, "handle valid: %3d %9s %10s\n",
                                h.fd, h.namemap->omsname,
                                h.namemap->linuxname);
                }
        }
        h.namemap = namemap - 1;
        if (validate_handle(&h)) {
                fprintf(stderr, "OOPS:  handle with bad namemap index ok\n");
                return;
        } else {
                fprintf(stderr,"handle with negative namemap index invalid\n");
        }
        fudge = (unsigned long)(namemap + 4) + 1;
        h.namemap = (struct namemap*) fudge;
        if (validate_handle(&h)) {
                fprintf(stderr, "OOPS:  handle with non-integer namemap "
                        "index ok\n");
                return;
        } else {
                fprintf(stderr, "handle with non-integer namemap index "
                        "invalid\n");
        }
}
